from typing import Optional, Sequence, Literal, Tuple
from sklearn.cluster import KMeans  # type: ignore
import torch

from .stochastic_matrix import EmissionMatrix
from .utils import validate_means, validate_covars, fill_covars

class CategoricalEmissions:
    """
    Categorical emission distribution for HMMs.

    Parameters:
    ----------
    n_dims (int):
        Number of hidden states in the model.
    n_emissions (int):
        Number of emissions in the model.
    init_params (bool):
        Whether to initialize the emission parameters.
    alpha (float):
        Dirichlet concentration parameter for the prior over emission probabilities.
    seed (int):
        Random seed for reproducibility.
    device (torch.device):
        Device on which to fit the model.

    Attributes:
    ----------
    emission_matrix (EmissionMatrix):
        Emission matrix representing the categorical distribution.
    """

    def __init__(self,
                n_dims: int,
                n_emissions: int,
                init_params: bool = True,
                alpha: float = 1.0,
                seed: Optional[int] = None,
                device: Optional[torch.device] = None):
            
        self.n_dims = n_dims
        self.n_emissions = n_emissions
        self.device = device

        if init_params:
            self._emission_matrix = self.sample_emissions_params(alpha, seed)  

    @property
    def emission_matrix(self) -> EmissionMatrix:
        return self._emission_matrix
    
    @emission_matrix.setter
    def emission_matrix(self, matrix):
        assert (self.n_dims,self.n_emissions) == matrix.shape, f'Expected matrix shape {(self.n_dims,self.n_emissions)} but got {matrix.shape}'
        if isinstance(matrix, EmissionMatrix):
            self._emission_matrix = matrix
        elif isinstance(matrix, torch.Tensor):
            self._emission_matrix = EmissionMatrix(n_states=self.n_dims, 
                                                   n_emissions=self.n_emissions, 
                                                   matrix=matrix,
                                                   device=self.device)
        else:
            raise NotImplementedError('Matrix type not supported')

    def sample_emissions_params(self, 
                                alpha: float, 
                                seed: Optional[int]=None) -> EmissionMatrix:
        return EmissionMatrix(n_states=self.n_dims,
                            n_emissions=self.n_emissions,
                            rand_seed=seed,
                            alpha=alpha,
                            device=self.device)

    def map_emission(self, x: torch.Tensor) -> torch.Tensor:
        return self._emission_matrix[:,x].squeeze()

    def compute_emprobs(self, 
                        X: Sequence[torch.Tensor],
                        gamma: torch.Tensor) -> torch.Tensor:  
        """Compute the emission probabilities for each hidden state."""
        emission_mat = torch.zeros(size=(self.n_dims, self.n_emissions),
                                   dtype=torch.float64,
                                   device=self.device)

        for i, seq in enumerate(X):
            seq_size = seq.size(dim=0)
            
            # Creating masks for each emission and summing up gamma values
            masks = (seq.view(1, 1, seq_size) == torch.arange(self.n_emissions, device=self.device).view(1, -1, 1))
            gamma_i = gamma[i, :, :seq_size].unsqueeze(1)
            emission_mat += (gamma_i * masks).sum(dim=2)
                    
        # Normalizing the emission matrix
        emission_mat /= gamma.sum(dim=0).sum(dim=1, keepdim=True)

        return emission_mat.log()

    def update_emissions_params(self, X: Sequence[torch.Tensor], gamma: torch.Tensor):
        self._emission_matrix.matrix.copy_(self.compute_emprobs(X, gamma))
        

class GaussianEmissions:

    """
    Gaussian Distribution for HMM emissions.    
    
    Parameters:
    ----------
    n_dims (int):
        Number of mixtures in the model. This is equal to the number of hidden states in the HMM.
    n_components (int):
        Number of components in the mixture model.
    n_features (int):
        Number of features in the data.
    init_weights (bool):
        Whether to initialize the mixture weights prior to fitting.
    init_dist ():
        Distribution to use for initializing the mixture weights.
    alpha (float):
        Dirichlet concentration parameter for the prior over mixture weights.
    k_means (bool):
        Whether to initialize the mixture means using K-Means clustering.
    min_covar (float):
        Minimum covariance for the mixture components.
    covariance_type (COVAR_TYPES):
        Type of covariance matrix to use for the mixture components. One of 'spherical', 'tied', 'diag', 'full'.
    dims (Sequence[str]):
        Names of the mixtures in the model.
    seed (int):
        Random seed for reproducibility.
    device (torch.device):
        Device to use for computations.
    """

    COVAR_TYPES_HINT = Literal['spherical', 'tied', 'diag', 'full']

    def __init__(self, 
                 n_dims: int,
                 n_features: int,
                 params_init: bool = True,
                 covariance_type: COVAR_TYPES_HINT = 'full',
                 min_covar: float = 1e-3,
                 random_state: Optional[int] = None,
                 device: Optional[torch.device] = None):
        
        self.n_dims = n_dims
        self.n_features = n_features
        self.device = device
        self.min_covar = min_covar
        self.covariance_type = covariance_type

        if params_init:
            self._means, self._covs = self.sample_emissions_params(random_state)
            
    @property
    def means(self) -> torch.Tensor:
        return self._means
    
    @means.setter
    def means(self, means: torch.Tensor):
        valid_means = validate_means(means, self.n_dims, self.n_features)
        self._means = valid_means.to(self.device)

    @property
    def covs(self) -> torch.Tensor:
        return self._covs

    @covs.setter
    def covs(self, new_covars: torch.Tensor):
        """Setter function for the covariance matrices."""
        valid_covars = validate_covars(new_covars, self.covariance_type, self.n_dims, self.n_features)
        self._covs = fill_covars(valid_covars, self.covariance_type, self.n_dims, self.n_features).to(self.device)

    @property
    def pdf(self):
        return torch.distributions.MultivariateNormal(loc=self.means, 
                                                      covariance_matrix=self.covs)
    
    def map_emission(self, x: torch.Tensor) -> torch.Tensor:
        return self.pdf.log_prob(x)
    
    @property
    def params(self):
        return {'means': self.means, 
                'covs': self.covs}
    
    def sample_emissions_params(self,
                                seed: Optional[int] = None) -> Tuple[torch.Tensor, torch.Tensor]:
        
        means = torch.zeros(size=(self.n_dims, self.n_features), 
                                    dtype=torch.float64, 
                                    device=self.device)
            
        covs = torch.eye(n=self.n_features, 
                                dtype=torch.float64,
                                device=self.device).expand((self.n_dims, self.n_features, self.n_features)).clone()
        # self._k_means = KMeans(n_clusters=(self.n_dims), 
        #                        random_state=seed, 
        #                        n_init="auto").fit(X)
        # torch.from_numpy(self._k_means.cluster_centers_).reshape(self.means.shape)
        return means, covs
    
    def compute_means(self, 
                      X: Sequence[torch.Tensor], 
                      gamma: torch.Tensor) -> torch.Tensor:
        """Compute the means for each hidden state"""
        new_mean = torch.zeros(size=(self.n_dims, self.n_features), 
                                dtype=torch.float64, 
                                device=self.device)
        
        for i, seq in enumerate(X):
            gamma_seq = gamma[i,:,:seq.size(dim=0)].unsqueeze(dim=-1)
            weighted_seq = gamma_seq.expand((-1, -1, self.n_features)) * seq
            new_mean += weighted_seq.sum(dim=-2)
            
        new_mean /= gamma.sum(dim=0).sum(dim=-1, keepdim=True)

        return new_mean

    def compute_covs(self, 
                     X: Sequence[torch.Tensor],
                     gamma: torch.Tensor) -> torch.Tensor:
        """Compute the covariances for each component."""
        new_covs = torch.zeros(size=(self.n_dims,self.n_features, self.n_features), 
                           dtype=torch.float64, 
                           device=self.device)

        for i, seq in enumerate(X):
            diff = seq.unsqueeze(0).expand(self.n_dims, -1, -1) - self.means.unsqueeze(-2)
            product = diff.unsqueeze(-1) @ diff.unsqueeze(-2)
            
            weighted_product = gamma[i, :, :seq.size(0)].unsqueeze(-1).unsqueeze(-1) * product
            new_covs += weighted_product.sum(dim=-3)

        new_covs /= gamma.sum(dim=0).sum(dim=-1, keepdim=True).unsqueeze(-1)
        new_covs += self.min_covar * torch.eye(self.n_features, device=self.device)

        return new_covs
    
    def update_emissions_params(self, X: Sequence[torch.Tensor], posterior: torch.Tensor):
        self._means.copy_(self.compute_means(X, posterior))
        self._covs.copy_(self.compute_covs(X, posterior))

class BernoulliEmissions: ...
class GammaEmissions: ...
class InverseGammaEmissions: ...
class InverseWishartEmissions: ...
class StudentTEmissions: ...
class WishartEmissions: ...
class MultinomialEmissions: ...
class ExponentialEmissions: ...
class weibullEmissions: ...
class BetaEmissions: ...
class NegativeBinomialEmissions: ...
class BinomialEmissions: ...
class PoissonEmissions: ...


from typing import Optional, Dict, Literal, Sequence
from abc import ABC, abstractproperty, abstractmethod

import torch
from torch.distributions import Distribution, MixtureSameFamily # type: ignore
from sklearn.cluster import KMeans # type: ignore
from stochastic_matrix import WeightsMatrix # type: ignore
from utils import validate_means, validate_covars, fill_covars # type: ignore


class MixtureEmissions(ABC):
    """
    Mixture model for HMM emissions. This class is an abstract base class for Gaussian, Poisson and other mixture models.
    """

    def __init__(self, 
                 n_dims: int,
                 n_components: int,
                 n_features: int,
                 init_weights: bool = True,
                 alpha: float = 1.0,
                 seed: Optional[int] = None,
                 device: Optional[torch.device] = None):
        
        self.n_components = n_components
        self.n_dims = n_dims
        self.n_features = n_features
        self.device = device
        
        if init_weights:
            self._weights = self.sample_weights(alpha, seed)

    @property
    def weights(self) -> WeightsMatrix:
        try:
            return self._weights
        except AttributeError:
            raise AttributeError('Weights are not initialized')

    @weights.setter
    def weights(self, vector):
        assert (self.n_dims, self.n_components) == vector.shape, 'Matrix dimensions differ from HMM model'
        if isinstance(vector, WeightsMatrix):
            self._weights = vector
        elif isinstance(vector, torch.Tensor):
            self._weights = WeightsMatrix(n_states=self.n_dims, 
                                          n_components=self.n_components,
                                          matrix=vector,
                                          device=self.device)
        else:
            raise NotImplementedError(f'Expected torch Tensor or WeightsMatrix object, got {type(vector)}')
        
    @property
    def mixture_pdf(self) -> MixtureSameFamily:
        """Return the emission distribution for Gaussian Mixture Distribution."""
        return MixtureSameFamily(mixture_distribution = self.weights._matrix,
                                 component_distribution = self.pdf)

    @abstractproperty
    def pdf(self) -> Distribution:
        """Return the emission distribution of Mixture."""
        pass

    @abstractproperty
    def params(self) -> Dict[str, torch.Tensor]:
        """Return the parameters of the Mixture."""
        pass

    @abstractmethod
    def update_emissions_params(self, X: Sequence[torch.Tensor], posterior: torch.Tensor) -> None:
        """Update the parameters of the Mixture."""
        pass 
    
    def sample_weights(self, 
                       alpha: float, 
                       seed: Optional[int]) -> WeightsMatrix:
        """Sample the weights for the mixture."""
        return WeightsMatrix(n_states=self.n_dims,
                            n_components=self.n_components,
                            rand_seed=seed,
                            alpha=alpha,
                            device=self.device)

    def map_emission(self, 
                     emission: torch.Tensor) -> torch.Tensor:
        return torch.logsumexp(self.weights + self.pdf.log_prob(emission), dim=1)
    
    def compute_weights(self, gamma: torch.Tensor) -> torch.Tensor:    
        gamma_reduced = gamma.sum(0).sum(-1)
        weights_exp = gamma_reduced / gamma_reduced.sum(dim=1, keepdim=True)
        return weights_exp.log()
    
    def compute_responsibilities(self, 
                                 X: torch.Tensor) -> torch.Tensor:
        """Compute the responsibilities for each component."""
        n_observations = X.size(dim=0)
        log_responsibilities = torch.zeros(size=(self.n_dims,self.n_components,n_observations), 
                                           dtype=torch.float64, 
                                           device=self.device)

        for t in range(n_observations):
            comps_resp = self.weights.matrix + self.pdf.log_prob(X[t])
            log_responsibilities[:,:,t].add_(comps_resp - comps_resp.logsumexp(dim=1, keepdim=True))
        
        return log_responsibilities
    

class GaussianMixtureEmissions(MixtureEmissions):
    """
    Gaussian Mixture Model for HMM emissions.    
    
    Parameters:
    ----------
    n_dims (int):
        Number of mixtures in the model. This is equal to the number of hidden states in the HMM.
    n_components (int):
        Number of components in the mixture model.
    n_features (int):
        Number of features in the data.
    init_weights (bool):
        Whether to initialize the mixture weights prior to fitting.
    init_dist ():
        Distribution to use for initializing the mixture weights.
    alpha (float):
        Dirichlet concentration parameter for the prior over mixture weights.
    k_means (bool):
        Whether to initialize the mixture means using K-Means clustering.
    min_covar (float):
        Minimum covariance for the mixture components.
    covariance_type (COVAR_TYPES):
        Type of covariance matrix to use for the mixture components. One of 'spherical', 'tied', 'diag', 'full'.
    dims (Sequence[str]):
        Names of the mixtures in the model.
    seed (int):
        Random seed for reproducibility.
    device (torch.device):
        Device to use for computations.
    """

    COVAR_TYPES_HINT = Literal['spherical', 'tied', 'diag', 'full']

    def __init__(self, 
                 n_dims: int,
                 n_components: int,
                 n_features: int,
                 params_init: bool = True,
                 alpha: float = 1.0,
                 covariance_type: COVAR_TYPES_HINT = 'full',
                 min_covar: float = 1e-3,
                 seed: Optional[int] = None,
                 device: Optional[torch.device] = None):
        
        self.min_covar = min_covar
        self.covariance_type = covariance_type
        
        MixtureEmissions.__init__(self,n_dims,n_components,n_features,params_init,alpha,seed,device)
        
        self._means = torch.zeros(size=(self.n_dims, self.n_components, self.n_features), 
                                  dtype=torch.float64, 
                                  device=self.device)
        
        self._covs = torch.eye(n=self.n_features, 
                               dtype=torch.float64,
                               device=self.device).expand((self.n_dims, self.n_components, self.n_features, self.n_features)).clone()
            
    @property
    def means(self) -> torch.Tensor:
        return self._means
    
    @means.setter
    def means(self, means: torch.Tensor):
        valid_means = validate_means(means, self.n_dims, self.n_features, self.n_components)
        self._means = valid_means.to(self.device)

    @property
    def covs(self) -> torch.Tensor:
        return self._covs

    @covs.setter
    def covs(self, new_covars: torch.Tensor):
        """Setter function for the covariance matrices."""
        valid_covars = validate_covars(new_covars, self.covariance_type, self.n_dims, self.n_features, self.n_components)
        self._covs = fill_covars(valid_covars, self.covariance_type, self.n_dims, self.n_features, self.n_components).to(self.device)

    @property
    def pdf(self):
        return torch.distributions.MultivariateNormal(loc=self.means, 
                                                      covariance_matrix=self.covs)
    
    @property
    def params(self):
        return {'weights': self.weights.matrix,
                'means': self.means, 
                'covs': self.covs}
    
    def sample_emissions_params(self, 
                                X: torch.Tensor,
                                seed: Optional[int] = None) -> torch.Tensor:
        self._k_means = KMeans(n_clusters=(self.n_dims*self.n_components), 
                               random_state=seed, 
                               n_init="auto").fit(X)
        return torch.from_numpy(self._k_means.cluster_centers_).reshape(self.means.shape)
    
    def compute_means(self, 
                      X: Sequence[torch.Tensor], 
                      responsibilities: torch.Tensor) -> torch.Tensor:
        """Compute the means for each component."""
        new_mean = torch.zeros(size=(self.n_dims,self.n_components,self.n_features), 
                                dtype=torch.float64, 
                                device=self.device)
        
        for i, seq in enumerate(X):
            gamma_seq = responsibilities[i,:,:seq.size(dim=0)].unsqueeze(dim=-1)
            weighted_seq = gamma_seq.expand((-1,self.n_components,-1,self.n_features)) * seq
            new_mean += weighted_seq.sum(dim=-2)
            
        new_mean /= responsibilities.sum(dim=0).sum(dim=-1, keepdim=True)

        return new_mean
    
    def compute_covs(self, 
                     X: Sequence[torch.Tensor],
                     responsibilities: torch.Tensor) -> torch.Tensor:
        """Compute the covariances for each component."""
        new_covs = torch.zeros(size=(self.n_dims,self.n_components,self.n_features,self.n_features), 
                           dtype=torch.float64, 
                           device=self.device)

        for i, seq in enumerate(X):
            diff = seq.expand(self.n_dims,self.n_components,-1, -1) - self.means.unsqueeze(-2)
            product = diff.unsqueeze(-1) @ diff.unsqueeze(-2)
            weighted_product = responsibilities[i,:,:seq.size(0)].unsqueeze(-1).unsqueeze(-1) * product
            new_covs += weighted_product.sum(dim=-3)

        new_covs /= responsibilities.sum(dim=0).sum(dim=-1, keepdim=True).unsqueeze(-1)
        new_covs += self.min_covar * torch.eye(self.n_features, device=self.device)
        
        return new_covs
    
    def update_emissions_params(self, X, posterior):
        self.weights.matrix.copy_(self.compute_weights(posterior))
        self._means.copy_(self.compute_means(X, posterior)) 
        self._covs.copy_(self.compute_covs(X, posterior))


from abc import ABC, abstractmethod, abstractproperty
from typing import Optional, Union, Sequence, List, TypeVar

import torch
import numpy as np
from torch.distributions import Categorical
from utils import print_table, states_names, init_matrix # type: ignore

class StochasticMatrix(ABC): ...
class EmissionMatrix(StochasticMatrix): ...
class TransitionMatrix(StochasticMatrix): ...
class ProbabilityVector(StochasticMatrix): ...
class DurationMatrix(StochasticMatrix): ...
class WeightsMatrix(StochasticMatrix): ...

MAT_OPS = TypeVar('MAT_OPS', bound=Union[StochasticMatrix, np.ndarray, torch.Tensor])
SAMP_DISTS = frozenset(("Dirichlet", "Uniform"))

torch.set_printoptions(precision=4, profile="full")

class StochasticMatrix(ABC):
    """ 
    Stochastic Matrix
    ----------
    This class is used to represent a stochastic matrix. It is a wrapper around a torch.Tensor with a
    torch.distributions.Categorical distribution.

    Parameters:
    ----------
    matrix (Categorical):
        The stochastic matrix.
    seed (Optional[int]):
        Random seed to use for reproducible results.
    
    Attributes:
    ----------
    matrix (torch.Tensor):
        The stochastic matrix.
    seed (Optional[int]):
        Random seed to use for reproducible results.
    shape (torch.Size):
        Shape of the stochastic matrix.
    """

    def __init__(self,
                 matrix: torch.Tensor,
                 device: Optional[torch.device] = None,
                 seed: Optional[int] = None):

        if device is None:
            self.device = torch.device('cuda' if torch.cuda.is_available() else 'cpu')
        else:
            self.device = device
        
        self._matrix = matrix.to(self.device)
        self.seed = seed

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}(shape={self.matrix.shape})"

    @abstractmethod
    def __str__(self) -> str:
        pass

    @abstractproperty
    def view(self):
        pass

    @property
    def matrix(self) -> torch.Tensor:
        return self._matrix

    @matrix.setter
    def matrix(self, matrix: Union[torch.Tensor, np.ndarray]):
        assert self.shape == matrix.shape(), "Matrix must be the same shape as the original"
        self._matrix = Categorical(logits=matrix, validate_args=True)
        
    def __mul__(self, other: MAT_OPS) -> torch.Tensor:
        if isinstance(other, StochasticMatrix):
            return self.matrix * other.matrix
        elif isinstance(other, torch.Tensor):
            return self.matrix * other
        elif isinstance(other, np.ndarray):
            return self.matrix * torch.from_numpy(other)
        else:
            raise NotImplementedError(f'Expected one of {MAT_OPS}, got {type(other)}')
        
    def __rmul__(self, other: MAT_OPS) -> torch.Tensor:
        return self.__mul__(other)
    
    def __matmul__(self, other: MAT_OPS) -> torch.Tensor:
        if isinstance(other, StochasticMatrix):
            return self.matrix @ other.matrix
        elif isinstance(other, torch.Tensor):
            return self.matrix @ other
        elif isinstance(other, np.ndarray):
            return self.matrix @ torch.from_numpy(other)
        else:
            raise NotImplementedError(f'Expected one of {MAT_OPS}, got {type(other)}')
    
    def __rmatmul__(self, other: MAT_OPS) -> torch.Tensor:
        return self.__matmul__(other)

    def __truediv__(self, other: MAT_OPS) -> torch.Tensor:
        if isinstance(other, StochasticMatrix):
            return self.matrix / other.matrix
        elif isinstance(other, torch.Tensor):
            return self.matrix / other
        elif isinstance(other, np.ndarray):
            return self.matrix / torch.from_numpy(other)
        else:
            raise NotImplementedError(f'Expected one of {MAT_OPS}, got {type(other)}')
    
    def __rtruediv__(self, other: MAT_OPS) -> torch.Tensor:
        return self.__truediv__(other)

    def __add__(self, other: MAT_OPS) -> torch.Tensor:
        if isinstance(other, StochasticMatrix):
            return self.matrix + other.matrix
        elif isinstance(other, torch.Tensor):
            return self.matrix + other
        elif isinstance(other, np.ndarray):
            return self.matrix + torch.from_numpy(other)
        else:
            raise NotImplementedError(f'Expected one of {MAT_OPS}, got {type(other)}')
        
    def __radd__(self, other: MAT_OPS) -> torch.Tensor:
        return self.__add__(other)
    
    def __sub__(self, other: MAT_OPS) -> torch.Tensor:
        if isinstance(other, StochasticMatrix):
            return self.matrix - other.matrix
        elif isinstance(other, torch.Tensor):
            return self.matrix - other
        elif isinstance(other, np.ndarray):
            return self.matrix - torch.from_numpy(other)
        else:
            raise NotImplementedError(f'Expected one of {MAT_OPS}, got {type(other)}')
        
    def __rsub__(self, other: MAT_OPS) -> torch.Tensor:
        return self.__sub__(other)

    def __eq__(self, other: MAT_OPS) -> bool:
        if isinstance(other, StochasticMatrix):
            return torch.equal(self.matrix, other.matrix)
        elif isinstance(other, torch.Tensor):
            return torch.equal(self.matrix, other)
        elif isinstance(other, np.ndarray):
            return torch.equal(self.matrix, torch.from_numpy(other))
        else:
            raise NotImplementedError(f'Expected one of {MAT_OPS}, got {type(other)}')
    
    def __len__(self) -> int:
        return self.matrix.numel()
    
    def __getitem__(self, idx) -> torch.Tensor:
        if isinstance(idx, tuple):
            row_idx, col_idx = idx
            if isinstance(row_idx, slice) or isinstance(col_idx, slice):
                return self.matrix[row_idx, col_idx]
            else:
                return self.matrix[row_idx, col_idx]
        else:
            return self.matrix[idx]
    
    def __setitem__(self, row_idx, col_idx, value):
        self.matrix[row_idx, col_idx] = value


class EmissionMatrix(StochasticMatrix):
    """
    Emission Matrix
    ---------------
    A Class representing the Emission matrix of a finite Markov chain.
    """

    def __init__(self,
                 n_states: int,
                 n_emissions: int,
                 matrix: Optional[torch.Tensor] = None,
                 states: Optional[Sequence[str]]= None,
                 emissions: Optional[Sequence[str]]= None,
                 rand_seed: Optional[int] = None,
                 alpha: float = 1.0,
                 device: Optional[torch.device] = None):
        
        self._n_states = n_states
        self._n_emissions = n_emissions
        self.states = states_names(n=n_states, state_type='state') if states is None else states
        self.emissions = states_names(n=n_emissions, state_type='emission') if emissions is None else emissions

        if matrix is None:
            mat_size = (n_states, n_emissions)
            matrix = torch.log(init_matrix(prior = alpha, target_size=mat_size))

        super().__init__(matrix=matrix,
                         seed=rand_seed,
                         device=device)
        
    def __str__(self) -> str:
        return f"EmissionMatrix(n_states = {self.n_states}, n_emissions = {self.n_emissions})"

    def view(self, log_scale: bool = False):
        vals = self.matrix.exp() if not log_scale else self.matrix
        vals = torch.round(vals, decimals=3).tolist()
        
        trans = [[self.states[idx]] + row for idx, row in enumerate(vals)]

        print_table(title='Emission Matrix',
                    header=['State\Emission'] + self.emissions, 
                    rows=trans)

    @property
    def n_states(self) -> int:
        return self._n_states

    @property
    def n_emissions(self) -> int:
        return self._n_emissions

    @property
    def states(self) -> Optional[List[str]]:
        return self._states

    @states.setter
    def states(self, states: Sequence[str]):
        unique_names = len(set(states))

        if unique_names != self.n_states:
            raise ValueError(f'Expected {self.n_states} unique states, got {unique_names}')
        else:
            self._states = states

    @property
    def emissions(self) -> Optional[List[str]]:
        return self._emissions

    @emissions.setter
    def emissions(self, emissions: Sequence[str]):
        unique_names = len(set(emissions))
        
        if unique_names != self.n_emissions:
            raise ValueError(f'Expected {self.n_emissions} unique emissions, got {unique_names}')
        else:
            self._emissions = emissions

class TransitionMatrix(StochasticMatrix):
    """
    Transition Matrix
    --------
    A Class representing the Transition matrix of a finite Markov chain.

    Parameters
    ----------
    n_states : int
        Number of states in the system.
    matrix : Optional[torch.Tensor], optional
        An initial probability distribution matrix, by default `None`.
    states : Optional[List[str]], optional
        A list of names/labels for each state in the system, by default `None`.
    rand_seed : Optional[int], optional
        The seed for the random number generator, by default `None`.
    init_dist : Optional[SAMPLING_DISTRIBUTIONS], optional
        The distribution to use when initializing the matrix if not provided, by default `'Uniform'`.
    alpha : Optional[float], optional
        The prior to use when initializing the matrix if not provided, by default `1.0`.

    Note
    ----
    If `matrix` is provided, it will be used as the initial probability distribution. 
    If `matrix` is not provided, a distribution specified by `init_dist` will be used to initialize the matrix.
    """

    def __init__(self,
                 n_states: int,
                 matrix: Optional[torch.Tensor] = None,
                 states: Optional[Sequence[str]]= None,
                 rand_seed: Optional[int] = None,
                 alpha: float = 1.0,
                 semi_markov: bool = False,
                 device: Optional[torch.device] = None):
        
        self._n_states = n_states
        self.semi_markov = semi_markov
        self.states = states_names(n=n_states, state_type='state') if states is None else states

        if matrix is None:
            mat_size = (n_states, n_states)
            matrix = torch.log(init_matrix(prior = alpha, target_size=mat_size, semi=semi_markov))

        super().__init__(matrix=matrix,
                         seed=rand_seed,
                         device=device)

    def __str__(self) -> str:
        return f"TransitionMatrix(n_states = {self.n_states})"

    def view(self, log_scale: bool = False):
        vals = self.matrix.exp() if not log_scale else self.matrix
        vals = torch.round(vals, decimals=3).tolist()
        
        trans = [[self.states[idx]] + row for idx, row in enumerate(vals)]

        print_table(title='Transition Matrix',
                    header=['State\State'] + self.states, 
                    rows=trans)

    @property
    def n_states(self) -> int:
        return self._n_states

    @property
    def states(self) -> Optional[List[str]]:
        return self._states

    @states.setter
    def states(self, states: list):
        if len(set(states)) != self.n_states:
            raise ValueError(f'Expected {self.n_states} states, got {len(set(states))}')
        else:
            self._states = states


class DurationMatrix(StochasticMatrix):
    """
    Duration Matrix
    ---------------
    A Class representing the Duration matrix of a semi-Markov chain.
    """

    def __init__(self,
                 n_states: int,
                 max_dur: int,
                 matrix: Optional[torch.Tensor] = None,
                 states: Optional[Sequence[str]]= None,
                 durations: Optional[Sequence[str]]= None,
                 rand_seed: Optional[int] = None,
                 alpha: float = 1.0,
                 device: Optional[torch.device] = None):
        
        self._n_states = n_states
        self._max_dur = max_dur
        self.states = states_names(n=n_states, state_type='state') if states is None else states
        self.durations = states_names(n=max_dur, state_type='duration') if durations is None else durations

        if matrix is None:
            mat_size = (n_states, max_dur)
            matrix = torch.log(init_matrix(prior = alpha, target_size=mat_size))

        super().__init__(matrix=matrix,
                         seed=rand_seed,
                         device=device)
        
    def __str__(self) -> str:
        return f"DurationMatrix(states = {self.n_states}, max_duration = {self.max_duration})"

    def view(self, log_scale: bool = False):
        vals = self.matrix.exp() if not log_scale else self.matrix
        vals = torch.round(vals, decimals=3).tolist()
        
        trans = [[self.states[idx]] + row for idx, row in enumerate(vals)]

        print_table(title='Duration Matrix',
                    header=['State\Emission'] + self.durations, 
                    rows=trans)

    @property
    def n_states(self) -> int:
        return self._n_states

    @property
    def durations(self) -> Optional[List[str]]:
        return self._durations

    @durations.setter
    def durations(self, durations: list):
        if len(set(durations)) != self.max_duration:
            raise ValueError(f'Expected {self.max_duration} unique durations, got {len(set(durations))}')
        else:
            self._durations = durations

    @property
    def states(self) -> Optional[List[str]]:
        return self._states

    @states.setter
    def states(self, states: list):
        if len(set(states)) != self.n_states:
            raise ValueError(f'Expected {self.n_states} unique states, got {len(set(states))}')
        else:
            self._states = states

    @property
    def max_duration(self) -> int:
        return self._max_dur


class WeightsMatrix(StochasticMatrix):
    """
    Weights Matrix
    ---------------
    A Class representing the Weights matrix of a Gaussian Mixture HMM.
    """

    def __init__(self,
                 n_states: int,
                 n_components: int,
                 matrix: Optional[torch.Tensor] = None,
                 states: Optional[Sequence[str]] = None,
                 components: Optional[Sequence[str]] = None,
                 rand_seed: Optional[int] = None,
                 alpha: float = 1.0,
                 device: Optional[torch.device] = None):
        
        self._n_states = n_states
        self._n_components = n_components
        self.states = states_names(n=n_states, state_type='state') if states is None else states
        self.components = states_names(n=n_components, state_type='component') if components is None else components

        if matrix is None:
            mat_size = (n_states, n_components)
            matrix = torch.log(init_matrix(prior = alpha, target_size=mat_size))

        super().__init__(matrix=matrix,
                         seed=rand_seed,
                         device=device)
        
    def __str__(self) -> str:
        return f"WeightsMatrix(n_states = {self.n_states}, n_components = {self.n_components})"

    def view(self, log_scale: bool = False):
        vals = self.matrix.exp() if not log_scale else self.matrix
        vals = torch.round(vals, decimals=3).tolist()
        
        trans = [[self.states[idx]] + row for idx, row in enumerate(vals)]

        print_table(title='Weights Matrix',
                    header=['State\Component'] + self.components, 
                    rows=trans)

    @property
    def n_states(self) -> int:
        return self._n_states

    @property
    def n_components(self) -> int:
        return self._n_components

    @property
    def states(self) -> Optional[List[str]]:
        return self._states

    @states.setter
    def states(self, states: list):
        if len(set(states)) != self.n_states:
            raise ValueError(f'Expected {self.n_states} unique states, got {len(set(states))}')
        else:
            self._states = states

    @property
    def components(self) -> Optional[List[str]]:
        return self._components
    
    @components.setter
    def components(self, components: list):
        if len(set(components)) != self.n_components:
            raise ValueError(f'Expected {self.n_components} unique components, got {len(set(components))}')
        else:
            self._components = components


class ProbabilityVector(StochasticMatrix):
    """
    Probability Vector
    ------------------
    A class used to represent a probability vector.

    Parameters
    ----------
    n_states : int
        Number of states in the system.
    vector : Optional[torch.Tensor], optional
        An initial probability distribution vector, by default `None`.
    states : Optional[List[str]], optional
        A list of names/labels for each state in the system, by default `None`.
    rand_seed : Optional[int], optional
        The seed for the random number generator, by default `None`.
    init_dist : Optional[SAMPLING_DISTRIBUTIONS], optional
        The distribution to use when initializing the vector if not provided, by default `'Uniform'`.
    alpha : Optional[float], optional
        The prior to use when initializing the vector if not provided, by default `1.0`.

    Note
    ----
    If `vector` is provided, it will be used as the initial probability distribution. 
    If `vector` is not provided, a distribution specified by `init_dist` will be used to initialize the vector.
    """

    def __init__(self,
                 n_states: int,
                 vector: Optional[torch.Tensor] = None,
                 states: Optional[Sequence[str]]= None,
                 rand_seed: Optional[int] = None,
                 alpha: float = 1.0,
                 device: Optional[torch.device] = None):
        
        self._n_states = n_states
        self.states = states_names(n=n_states, state_type='state') if states is None else states

        if vector is None:
            mat_size = (n_states,)
            vector = torch.log(init_matrix(prior = alpha, target_size=mat_size))

        super().__init__(matrix=vector,
                         seed=rand_seed,
                         device=device)

    def __str__(self) -> str:
        return f"ProbabilityVector(n_states = {self.n_states})"
    
    def view(self, log_scale: bool = False):
        vals = self.matrix.exp() if not log_scale else self.matrix
        row_name = 'Log Prob' if log_scale else 'Prob'
        vals = torch.round(vals, decimals=3).tolist()
        
        print_table(title='Initial Distribution',
                    header=['State'] + self.states, 
                    rows=[[row_name] + vals])

    @property
    def n_states(self) -> int:
        return self._n_states

    @property
    def states(self) -> List[str]:
        return self._states

    @states.setter
    def states(self, states: list):
        if len(set(states)) == self.n_states:
            self._states = states
        else:
            raise ValueError(f'Expected {self.n_states} unique states, got {len(set(states))}')
        


"""
ChadHMM
======

Ultra Chad Implementation of Hidden Markov Models in Pytorch (available only to true sigma males)

But seriously this package needs you to help me make it better. I'm not a professional programmer, I'm just a guy who likes to code. 
If you have any suggestions, please let me know. I'm open to all ideas.
"""

# Set version
__version__ = '0.0.1'
__author__ = 'GarroshIcecream'
__license__ = 'MIT'

# Import HMM objects
from .hmm.CategoricalHMM import CategoricalHMM
from .hmm.GaussianHMM import GaussianHMM, GaussianMixtureHMM
from .hsmm.CategoricalHSMM import CategoricalHSMM
from .hsmm.GaussianHSMM import GaussianHSMM, GaussianMixtureHSMM